From 152b428272eeacf7487cc036721a9e12e63c85c6 Mon Sep 17 00:00:00 2001
From: Juan Jose Garcia Ripoll <jjgarcia@users.sourceforge.net>
Date: Sun, 25 Apr 2010 14:41:50 +0200
Subject: [PATCH 4/4] Teach CFFI to use C-INLINE when producing compiled code

Using the ECL extension EXT:WITH-BACKEND, CFFI can produce code that works
differently when using the interpreter than when using the lisp2C compiler.
This leads to more efficient code. This change is backwards compatible.
---
 src/cffi-ecl.lisp |   61 ++++++++++++++++++++++++++++++++++------------------
 1 files changed, 40 insertions(+), 21 deletions(-)

diff --git a/src/cffi-ecl.lisp b/src/cffi-ecl.lisp
index c137fdc..394b262 100644
--- a/src/cffi-ecl.lisp
+++ b/src/cffi-ecl.lisp
@@ -226,29 +226,48 @@ WITH-POINTER-TO-VECTOR-DATA."
 
 (defconstant +ecl-inline-codes+ "#0,#1,#2,#3,#4,#5,#6,#7,#8,#9,#a,#b,#c,#d,#e,#f,#g,#h,#i,#j,#k,#l,#m,#n,#o,#p,#q,#r,#s,#t,#u,#v,#w,#x,#y,#z")
 
+(defun c-inline-function-pointer-call (pointer types values return-type)
+  (when (stringp pointer)
+    (setf pointer `(%foreign-symbol-pointer ,pointer nil)))
+  `(ffi:c-inline
+    ,(list* pointer values)
+    ,(list* :pointer-void types) ,return-type
+    ,(with-output-to-string (s)
+       (let ((types (mapcar #'ecl-type->c-type types)))
+         ;; On AMD64, the following code only works with the extra
+         ;; argument ",...". If this is not present, functions
+         ;; like sprintf do not work
+         (format s "((~A (*)(~@[~{~A,~}...~]))(#0))(~A)"
+                 (ecl-type->c-type return-type) types
+                 (subseq +ecl-inline-codes+ 3
+                         (max 3 (+ 2 (* (length values) 3)))))))
+    :one-liner t :side-effects t))
+
+(defun dffi-function-pointer-call (pointer types values return-type)
+  (when (stringp pointer)
+    (setf pointer `(%foreign-symbol-pointer ,pointer nil)))
+  `(si:call-cfun ,pointer ,return-type (list ,@types) (list ,@values)))
+
+#.(cl:when (>= ext:+ecl-version-number+ 100402)
+    (cl:pushnew :ecl-with-backend cl:*features*)
+    cl:nil)
+
 (defun produce-function-pointer-call (pointer types values return-type)
-  #-dffi
-  (if (stringp pointer)
-      (produce-function-pointer-call
-       `(%foreign-symbol-pointer ,pointer nil) types values return-type)
-      `(ffi:c-inline
-        ,(list* pointer values)
-        ,(list* :pointer-void types) ,return-type
-        ,(with-output-to-string (s)
-           (let ((types (mapcar #'ecl-type->c-type types)))
-             ;; On AMD64, the following code only works with the extra
-             ;; argument ",...". If this is not present, functions
-             ;; like sprintf do not work
-             (format s "((~A (*)(~@[~{~A,~}...~]))(#0))(~A)"
-                     (ecl-type->c-type return-type) types
-                     (subseq +ecl-inline-codes+ 3
-                             (max 3 (+ 2 (* (length values) 3)))))))
-        :one-liner t :side-effects t))
-  #+dffi
+  #-ecl-with-backend
   (progn
-    (when (stringp pointer)
-      (setf pointer `(%foreign-symbol-pointer ,pointer nil)))
-    `(si:call-cfun ,pointer ,return-type (list ,@types) (list ,@values))))
+    #-dffi
+    (dffi-function-pointer-call pointer types values return-type)
+    #+dffi
+    (c-inline-function-pointer-call pointer types values return-type))
+  #+ecl-with-backend
+  `(ext:with-backend 
+     :bytecodes
+     #-dffi
+     (error "In interpreted code, attempted to call a foreign function~% ~A~%but ECL was built without support for that." pointer)
+     #+dffi
+     ,(dffi-function-pointer-call pointer types values return-type)
+     :c/c++
+     ,(c-inline-function-pointer-call pointer types values return-type)))
 
 
 (defun foreign-funcall-parse-args (args)
-- 
1.6.6.1

